"""fprime_gds.test_plugins:

A set of pytest-enabled tests that are used to test the plugin architecture of the fprime_gds. This file defines a
number of plugins, instantiates a non-singleton plugin instance, and registers the plugins to it. These steps are
provided via the `plugins` fixture to tests.

@author lestarch
"""

import os
import signal
import subprocess
import sys
import time
from typing import Dict, Tuple

import pytest

from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from fprime_gds.common.communication.adapters.ip import IpAdapter
from fprime_gds.common.communication.adapters.base import NoneAdapter
from fprime_gds.common.communication.adapters.uart import SerialAdapter
from fprime_gds.common.communication.framing import FramerDeframer, FpFramerDeframer
from fprime_gds.common.pipeline.standard import StandardPipeline
from fprime_gds.executables.cli import ParserBase, PluginArgumentParser
from fprime_gds.executables.apps import GdsFunction, GdsApp, GdsStandardApp
from fprime_gds.plugin.definitions import gds_plugin_implementation, gds_plugin
from fprime_gds.plugin.system import Plugins


class BadSuperClass(object):
    """A bad framing plugin (inappropriate parent)

    This plugin implementation will fail to register as a "framing" plugin because it does not inherit from
    FramerDeframer.
    """

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """Register a bad plugin"""
        return cls


class BadImplementation(FramerDeframer):
    """A bad framing plugin (inappropriate implementation)

    This plugin implementation will fail to register as a "framing" plugin because it does not implement all abstract
    methods required by the `FramerDeframer` class.
    """

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """Register a bad plugin"""
        return cls


class Good(FramerDeframer):
    """A good framing plugin

    This plugin implementation inherits from the correct class and implements abstract functions. Thus, this plugin will
    register correctly as a "framing" plugin.
    """

    def frame(self, data):
        """Required function: frame data"""
        pass

    def deframe(self, data, no_copy=False):
        """Required function: deframe data"""
        pass

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """Register a good plugin"""
        return cls


class GoodWithArgs(Good):
    """A good framing plugin including arguments

    This plugin implementation is the same "good" implementation with the addition of arguments requested from the
    plugin system. This implementation will take and store these arguments to prove the system.

    Additionally, this plugin implementation defines the `get_name` class method to provide a name for the plugin.
    """

    NAMED_PLUGIN_NAME = "good-with-args"

    def __init__(self, my_fancy_arg, fancy_2):
        """Fancy argument input processing"""
        self.my_fancy_arg = my_fancy_arg
        self.fancy_2 = fancy_2

    @classmethod
    def get_name(cls):
        """Return the name of this function"""
        return cls.NAMED_PLUGIN_NAME

    @classmethod
    def get_arguments(cls):
        """Return some argument specifications"""
        return {
            ("--my-fancy-arg",): {"type": str, "help": "Some help string"},
            ("--my-fancy-arg-with-dest",): {
                "dest": "fancy_2",
                "type": int,
                "help": "Some help string",
            },
        }

    @classmethod
    def check_arguments(cls, my_fancy_arg, fancy_2):
        """Check arguments to raise ValueError"""
        if fancy_2 < 0:
            raise ValueError("Must be positive")


class StartFunction(GdsFunction):
    """A plugin implementation that starts a function

    This plugin implementation creates a start-up function that will run on the invocation of the GDS.
    """

    def __init__(self, start_up_file):
        """Start-Up function file annotation"""
        self.start_up_function_file = start_up_file

    def run(self, parsed_args):
        """Run function"""
        with open(self.start_up_function_file, "a+") as file_handle:
            file_handle.write("ACK\n")

    @classmethod
    def get_name(cls):
        """Get name"""
        return "test-start-function"

    @classmethod
    def get_arguments(cls):
        """Get arguments"""
        return {
            ("--start-up-file",): {
                "type": str,
                "help": "File to run start-up function on",
                "required": True,
            }
        }

    @classmethod
    @gds_plugin_implementation
    def register_gds_function_plugin(cls):
        """Register a good plugin"""
        return cls


class StartApp(GdsApp):
    """A plugin implementation that starts a app in a separate process

    This plugin implementation creates a start-up application that will run in a separate process
    """

    def __init__(self, namespace, start_up_file):
        """Start-Up function file annotation"""
        super().__init__()
        self.start_up_app_file = start_up_file

    def get_process_invocation(self, parsed_args):
        """Process invocation"""
        return [sys.executable, __file__, self.start_up_app_file, "ACK-2"]

    @classmethod
    def get_name(cls):
        """Get name"""
        return "test-start-app"

    @classmethod
    def get_arguments(cls):
        """Get arguments"""
        return {
            ("--start-up-file",): {
                "type": str,
                "help": "File to run start-up function on",
                "required": True,
            }
        }

    @classmethod
    @gds_plugin_implementation
    def register_gds_app_plugin(cls):
        """Register a good plugin"""
        return cls


@gds_plugin(GdsStandardApp)
class StandardAppTester(GdsStandardApp):
    """A test plugin that uses the standard app functionality"""

    INIT_CALLED = False

    def __init__(
        self, test_arg, test_arg_with_default, start_up_file, **pipeline_arguments
    ):
        """Initialize the standard app tester"""
        super().__init__(**pipeline_arguments)
        self.test_arg = test_arg
        self.test_arg_with_default = test_arg_with_default
        self.start_up_file = start_up_file

    @classmethod
    def get_additional_arguments(cls) -> Dict[Tuple, Dict[str, str]]:
        """Function to provide additional command line arguments beyond the standard pipeline

        Override this function to provide additional arguments. The form of the arguments are the same as returned by
        standard plugins: a dictionary of tuple flags to argparse kwargs inputs.

        Return:
            dictionary of flag tuple to argparse kwargs
        """
        return {
            ("--test-arg",): {"type": str, "help": "Test argument", "required": True},
            ("--test-arg-with-default",): {
                "type": str,
                "help": "Test argument with default",
                "default": "test-default",
            },
            ("--start-up-file",): {
                "type": str,
                "help": "File to run start-up function on",
                "required": True,
            },
        }

    @classmethod
    def init(cls):
        """Allows standard application plugins to initialize before argument parsing is performed"""
        cls.INIT_CALLED = True

    def start(self, pipeline: StandardPipeline):
        """Start function to contain behavior based in standard pipeline"""
        with open(self.start_up_file, "a+") as file_handle:
            if self.INIT_CALLED:
                print("[TestStandardApp] init called", file=file_handle)
            print("[TestStandardApp] start called", file=file_handle)
            print("[TestStandardApp] test-arg", self.test_arg, file=file_handle)
            print(
                "[TestStandardApp] test-arg-with-default",
                self.test_arg_with_default,
                file=file_handle,
            )


@pytest.fixture()
def plugins():
    """Register test plugins as fixture for testing

    Create a `plugins` fixture that is a plugin system created to register specifically the test plugins on top of the
    installed and provided plugins.
    """
    system = Plugins(["communication", "framing"])
    system.register_plugin(BadSuperClass)
    system.register_plugin(BadImplementation)
    system.register_plugin(Good)
    system.register_plugin(GoodWithArgs)
    Plugins._singleton = system
    return system


@pytest.fixture()
def start_up(request):
    """Register test start up plugins as a fixture for testing"""
    parent_path = Path(__file__).parent
    extra_plugins, flags = request.param

    # Update the environment to side-load python
    environment = os.environ.copy()
    environment[Plugins.PLUGIN_ENVIRONMENT_VARIABLE] = extra_plugins
    environment["PYTHONPATH"] = f"{environment.get('PYTHONPATH', '')}:{parent_path}"
    with TemporaryDirectory() as temp_dir:
        # Command line arguments including a temporary directory for the ZMQ transport
        # sockets used in this test.
        command_arguments = [
            "fprime-gds",
            "-n",
            "--dictionary",
            str(parent_path / "sample" / "dictionary.xml"),
            "--zmq-transport",
            f"ipc://{temp_dir}/fprime-test-in",
            f"ipc://{temp_dir}/fprime-test-out",
            "-g",
            "none",
            "--disable-custom-data-handlers",
        ] + flags
        with NamedTemporaryFile(mode="w+", dir=temp_dir) as temp_file:
            assert "" == temp_file.read(), "Failed to read empty file"
            command_arguments += ["--start-up-file", temp_file.name]
            # Run subprocess for 3 seconds then kill the GDS with 5 seconds to shut down
            process = subprocess.Popen(command_arguments, env=environment)
            time.sleep(3)
            process.send_signal(signal.SIGINT)
            _ = process.communicate(None, 5)
            yield temp_file


def test_base_plugin(plugins):
    """Tests good framing plugins are returned"""
    plugin_options = plugins.get_plugins("framing")
    assert Good in [
        plugin.plugin_class for plugin in plugin_options
    ], "Good plugin not registered as expected"


def test_framing_builtin_plugins(plugins):
    """Tests good framing plugins are returned"""
    plugin_options = plugins.get_plugins("framing")
    assert FpFramerDeframer in [
        plugin.plugin_class for plugin in plugin_options
    ], "FpFramerDeframer plugin not registered as expected"


def test_communication_builtin_plugins(plugins):
    """Tests good framing plugins are returned"""
    plugin_options = plugins.get_plugins("communication")
    plugin_classes = [plugin.plugin_class for plugin in plugin_options]
    for expected in [IpAdapter, NoneAdapter, SerialAdapter]:
        assert (
            expected in plugin_classes
        ), f"{expected.__name__} plugin not registered as expected"


def test_plugin_categories(plugins):
    """Tests plugin categories"""
    plugin_categories = plugins.get_categories()
    assert sorted(plugin_categories) == sorted(
        ["communication", "framing"]
    ), "Detected plugin categories incorrect"


def test_plugin_validation(plugins):
    """Tests good framing plugins are returned"""
    plugin_options = plugins.get_plugins("framing")
    plugin_classes = [plugin.plugin_class for plugin in plugin_options]
    assert (
        BadSuperClass not in plugin_classes
    ), "Plugin with bad parent class not excluded as expected"
    assert (
        BadImplementation not in plugin_classes
    ), "Plugin with abstract implementation not excluded as expected"


def test_plugin_arguments(plugins):
    """Tests that arguments can be parsed and supplied to a plugin"""
    plugin_system = plugins
    a_string = "a_string"
    a_number = "201"
    to_parse = [
        "--framing",
        "good-with-args",
        "--my-fancy-arg",
        a_string,
        "--my-fancy-arg-with-dest",
        a_number,
    ]
    args, _ = ParserBase.parse_args(
        [
            PluginArgumentParser(plugin_system),
        ],
        arguments=to_parse,
    )
    assert (
        args.framing_selection == GoodWithArgs.NAMED_PLUGIN_NAME
    ), "Improper framing selection"
    instance = plugin_system.get_selected_class("framing")()
    assert isinstance(instance, GoodWithArgs), "Invalid instance created"
    assert instance.my_fancy_arg == a_string, "String argument did not bind"
    assert instance.fancy_2 == int(a_number), "Integer argument did not bind"


def test_plugin_check_arguments(plugins):
    """Tests that arguments are validated in plugins"""
    a_string = "a_string"
    a_number = "-20"
    to_parse = [
        "--framing",
        "good-with-args",
        "--my-fancy-arg",
        a_string,
        "--my-fancy-arg-with-dest",
        a_number,
    ]
    with pytest.raises(SystemExit):
        args, _ = ParserBase.parse_args(
            [
                PluginArgumentParser(plugins),
            ],
            arguments=to_parse,
        )


def test_plugin_decorator(plugins):
    """Test that the plugin decorator works with known good class"""

    @gds_plugin(FramerDeframer)
    class MyGood(FramerDeframer):
        def frame(self, data):
            pass

        def deframe(self, data, no_copy=False):
            pass

    plugins.register_plugin(MyGood)
    plugin_options = plugins.get_plugins("framing")
    assert MyGood in [
        plugin.plugin_class for plugin in plugin_options
    ], "MyGood plugin not registered as expected"


def test_plugin_decorator_without_class():
    """Test that the plugin decorator fails if a class is not provided"""
    with pytest.raises(Exception):

        @gds_plugin
        class MyGood(Good):
            pass


def test_plugin_decorator_class_mismatch():
    """Test that the plugin decorator fails if it is supplied the wrong plugin class"""
    with pytest.raises(Exception):

        @gds_plugin(StartApp)
        class MyGood(Good):  # Good is a FramerDeframer
            pass


def test_plugin_decorator_not_a_plugin():
    """Test that the plugin decorator fails if it is supplied a class which is not a plugin"""
    with pytest.raises(Exception):

        @gds_plugin(object)
        class MyGood(Good):  # Good is a FramerDeframer
            pass


@pytest.mark.parametrize("start_up", [(f"{__name__}:StartFunction", [])], indirect=True)
def test_start_function(start_up):
    """Test start-up functions"""
    assert "ACK\n" == start_up.read(), "Failed to read expected data"


@pytest.mark.parametrize(
    "start_up",
    [(f"{__name__}:StartFunction", ["--disable-test-start-function"])],
    indirect=True,
)
def test_disabled_start_function(start_up):
    """Test disabled start-up functions"""
    assert (
        "" == start_up.read()
    ), "Failed to read empty file: Function ran when disabled"


@pytest.mark.parametrize("start_up", [(f"{__name__}:StartApp", [])], indirect=True)
def test_start_app(start_up):
    """Test start-up functions"""
    assert "ACK-2\n" == start_up.read(), "Failed to read expected data"


@pytest.mark.parametrize(
    "start_up", [(f"{__name__}:StartApp", ["--disable-test-start-app"])], indirect=True
)
def test_disabled_start_app(start_up):
    """Test disabled start-up functions"""
    assert "" == start_up.read(), "Failed to read empty file"


@pytest.mark.parametrize(
    "start_up",
    [(f"{__name__}:StandardAppTester", ["--test-arg", "test"])],
    indirect=True,
)
def test_standard_app(start_up):
    """Test standard app functionality"""
    lines = [line.strip() for line in start_up.readlines()]
    expected_lines = [
        "[TestStandardApp] init called",
        "[TestStandardApp] start called",
        "[TestStandardApp] test-arg test",
        "[TestStandardApp] test-arg-with-default test-default",
    ]
    assert expected_lines[0] in lines, "Plugin not initialized"
    assert expected_lines[1] in lines, "Plugin not started"
    assert expected_lines[2] in lines, "Test argument not passed"
    assert expected_lines[3] in lines, "Test argument with default not passed"
    assert expected_lines == lines, "Ordering of lines not correct"


@pytest.mark.parametrize(
    "start_up",
    [
        (
            f"{__name__}:StandardAppTester",
            ["--test-arg", "test", "--test-arg-with-default", "test-non-default"],
        )
    ],
    indirect=True,
)
def test_standard_app_with_non_defaults(start_up):
    """Test standard app functionality"""
    lines = [line.strip() for line in start_up.readlines()]
    expected_lines = [
        "[TestStandardApp] init called",
        "[TestStandardApp] start called",
        "[TestStandardApp] test-arg test",
        "[TestStandardApp] test-arg-with-default test-non-default",
    ]
    assert expected_lines[0] in lines, "Plugin not initialized"
    assert expected_lines[1] in lines, "Plugin not started"
    assert expected_lines[2] in lines, "Test argument not passed"
    assert (
        expected_lines[3] in lines
    ), "Test argument with overridden default not passed"
    assert expected_lines == lines, "Ordering of lines not correct"


def main():
    """Run main entry point function for StartApp plugin

    The main function is run when the file is invoked not as a pytest set of test, but rather when it is run as part of
    the StartApp plugin. This main program writes the second argument to the file represented as the first.
    """
    with open(sys.argv[1], "a+") as file_handle:
        file_handle.write(f"{sys.argv[2]}\n")


if __name__ == "__main__":
    main()



"""
@brief Class to log raw binary input and output as well as telemetry and events
"""

import os

import fprime_gds.common.handlers
from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.cmd_data import CmdData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.data_types.pkt_data import PktData


class DataLogger(fprime_gds.common.handlers.DataHandler):
    def __init__(self, logdir, verbose=False, csv=False, prefix=""):

        self.logdir = logdir

        self.recv_file = f'{prefix}recv.bin'
        self.send_file = f'{prefix}sent.bin'
        self.telem_file = f'{prefix}channel.log'
        self.event_file = f'{prefix}event.log'
        self.command_file = f'{prefix}command.log'

        self.verbose = verbose
        self.csv = csv
        self.f_r = open(self.logdir + os.sep + self.recv_file, "ab+")
        self.f_s = open(self.logdir + os.sep + self.send_file, "ab+")
        self.f_telem = open(self.logdir + os.sep + self.telem_file, "a+")
        self.f_event = open(self.logdir + os.sep + self.event_file, "a+")
        self.f_command = open(self.logdir + os.sep + self.command_file, "a+")

    def __del__(self):
        self.f_r.close()
        self.f_s.close()
        self.f_telem.close()
        self.f_event.close()

    def data_callback(self, data, sender=None):
        if isinstance(data, (ChData, PktData)):
            self.f_telem.write(data.get_str(verbose=self.verbose, csv=self.csv) + "\n")
            self.f_telem.flush()

        if isinstance(data, EventData):
            self.f_event.write(data.get_str(verbose=self.verbose, csv=self.csv) + "\n")
            self.f_event.flush()

        if isinstance(data, CmdData):
            self.f_command.write(
                data.get_str(verbose=self.verbose, csv=self.csv) + "\n"
            )
            self.f_command.flush()

        if isinstance(data, (bytes, bytearray)):
            self.on_recv(data)

    def send(self, data, dest):
        """Send callback for the encoder

        Arguments:
            data {bin} -- binary data packet
            dest {string} -- where the data will be sent by the server
        """

        self.f_s.write(data)
        self.f_s.flush()

    # Some data was recvd
    def on_recv(self, data):
        """Data was received on the socket server

        Arguments:
            data {bin} --binary data string that was received
        """

        self.f_r.write(data)
        self.f_r.flush()



####
# run_deployment.py:
#
# Runs a deployment. Starts a GUI, a TCPServer, and the deployment application.
####
import os
import sys
import copy
import webbrowser

from fprime_gds.executables.cli import (
    BinaryDeployment,
    ConfigDrivenParser,
    CompositeParser,
    CommParser,
    GdsParser,
    ParserBase,
    StandardPipelineParser,
    PluginArgumentParser,
)
from fprime_gds.executables.utils import AppWrapperException, run_wrapped_application
from fprime_gds.plugin.system import Plugins

BASE_MODULE_ARGUMENTS = [sys.executable, "-u", "-m"]


def parse_args():
    """Parse command line arguments
    Gets an argument parsers to read the command line and process the arguments. Return
    the arguments in their namespace.

    :return: parsed argument namespace
    """
    # Get custom handlers for all executables we are running
    arg_handlers = [
        StandardPipelineParser,
        GdsParser,
        BinaryDeployment,
        CommParser,
        PluginArgumentParser,
    ]
    # Parse the arguments, and refine through all handlers
    args, parser = ConfigDrivenParser.parse_args(
        arg_handlers, "Run F prime deployment and GDS"
    )
    return args


def launch_process(cmd, logfile=None, name=None, env=None, launch_time=5):
    """
    Launch a child subprocess. This subprocess will allow the child to run outside of the memory context of Python.

    :param cmd: list of command arguments to run by handing to subprocess.
    :param logfile: (optional) place to redirect output to for purposes of logging. Default: None, screen.
    :param name: (optional) short name for printing messages.
    :param env: (optional) environment to run in. Allows for special environment contexts.
    :param launch_time: (optional) time to launch the process, before rendering an error.
    :return: running process
    """
    if name is None:
        name = str(cmd)
    print(f"[INFO] Ensuring {name} is stable for at least {launch_time} seconds")
    try:
        return run_wrapped_application(cmd, logfile, env, launch_time)
    except AppWrapperException as awe:
        print(f"[ERROR] {str(awe)}.", file=sys.stderr)
        try:
            if logfile is not None:
                with open(logfile) as file_handle:
                    for line in file_handle:
                        print(f"    [LOG] {line.strip()}", file=sys.stderr)
        except Exception:
            pass
        msg = f"Failed to run {name}"
        raise AppWrapperException(msg)


def launch_tts(parsed_args):
    """Launch the ThreadedTcpServer middleware application


    Args:
        parsed_args: parsed argument namespace
    Return:
        launched process
    """
    # Open log, and prepare to close it cleanly on exit
    tts_log = os.path.join(parsed_args.logs, "ThreadedTCP.log")
    # Launch the tcp server
    tts_cmd = BASE_MODULE_ARGUMENTS + [
        "fprime_gds.executables.tcpserver",
        "--port",
        str(parsed_args.tts_port),
        "--host",
        str(parsed_args.tts_addr),
    ]
    return launch_process(tts_cmd, logfile=tts_log, name="TCP Server")


def launch_html(parsed_args):
    """Launch the Flask application

    Args:
        parsed_args: parsed argument namespace
    Return:
        launched process
    """
    composite_parser = CompositeParser([StandardPipelineParser, ConfigDrivenParser])
    reproduced_arguments = StandardPipelineParser().reproduce_cli_args(parsed_args)
    if "--log-directly" not in reproduced_arguments:
        reproduced_arguments += ["--log-directly"]
    flask_env = os.environ.copy()
    flask_env.update(
        {
            "FLASK_APP": "fprime_gds.flask.app",
            "STANDARD_PIPELINE_ARGUMENTS": "|".join(reproduced_arguments),
            "SERVE_LOGS": "YES",
        }
    )
    gse_args = BASE_MODULE_ARGUMENTS + [
        "flask",
        "run",
        "--host",
        str(parsed_args.gui_addr),
        "--port",
        str(parsed_args.gui_port),
    ]
    ret = launch_process(gse_args, name="HTML GUI", env=flask_env, launch_time=2)
    ui_url = f"http://{str(parsed_args.gui_addr)}:{str(parsed_args.gui_port)}/"
    print(f"[INFO] Launched UI at: {ui_url}")
    webbrowser.open(
        ui_url,
        new=0,
        autoraise=True,
    )
    return ret


def launch_app(parsed_args):
    """Launch the raw application

    Args:
        parsed_args: parsed argument namespace
    Return:
        launched process
    """
    app_path = parsed_args.app
    logfile = os.path.join(parsed_args.logs, f"{app_path.name}.log")
    app_cmd = [
        app_path.absolute(),
        "-p",
        str(parsed_args.port),
        "-a",
        parsed_args.address,
    ]
    return launch_process(
        app_cmd, name=f"{app_path.name} Application", logfile=logfile, launch_time=1
    )


def launch_comm(parsed_args):
    """Launch the communication adapter process

    Args:
        parsed_args: parsed argument namespace
    Return:
        launched process
    """
    arguments = CommParser().reproduce_cli_args(parsed_args)
    arguments = (
        arguments + ["--log-directly"]
        if "--log-directly" not in arguments
        else arguments
    )
    app_cmd = BASE_MODULE_ARGUMENTS + ["fprime_gds.executables.comm"] + arguments
    return launch_process(
        app_cmd,
        name=f"comm[{parsed_args.communication_selection}] Application",
        launch_time=1,
    )


def launch_plugin(parsed_args, plugin_class_instance):
    """Launch a plugin instance"""
    plugin_name = getattr(
        plugin_class_instance,
        "get_name",
        lambda: plugin_class_instance.__class__.__name__,
    )()
    plugin_args = copy.deepcopy(parsed_args)
    # Set logging to use a subdirectory within the root logs directory
    plugin_logs = os.path.join(plugin_args.logs, plugin_name)
    os.mkdir(plugin_logs)
    plugin_args.logs = plugin_logs
    plugin_args.log_directly = True
    return launch_process(
        plugin_class_instance.get_process_invocation(plugin_args),
        name=f"{ plugin_name } Plugin App",
        launch_time=1,
    )


def main():
    """
    Main function used to launch processes.
    """
    parsed_args = parse_args()
    launchers = []

    # Launch middleware layer if not using ZMQ
    if not parsed_args.zmq:
        launchers.append(launch_tts)

    # Check if we are running with communications
    if parsed_args.communication_selection != "none":
        launchers.append(launch_comm)

    # Add app, if possible
    if parsed_args.app:
        if parsed_args.communication_selection == "ip":
            launchers.append(launch_app)
        else:
            print(
                "[WARNING] App cannot be auto-launched without IP adapter",
                file=sys.stderr,
            )

    # Launch the desired GUI package
    if parsed_args.gui == "html":
        launchers.append(launch_html)

    # Launch launchers and wait for the last app to finish
    try:
        procs = [launcher(parsed_args) for launcher in launchers]
        _ = [
            launch_plugin(parsed_args, cls(namespace=parsed_args))
            for cls in Plugins.system().get_feature_classes("gds_app")
        ]
        _ = [
            instance().run(parsed_args)
            for instance in Plugins.system().get_feature_classes("gds_function")
        ]

        print("[INFO] F prime is now running. CTRL-C to shutdown all components.")
        procs[-1].wait()
    except KeyboardInterrupt:
        print("[INFO] CTRL-C received. Exiting.")
    except Exception as exc:
        print(f"[INFO] Shutting down F prime due to error. {str(exc)}", file=sys.stderr)
        return 1
    # Processes are killed atexit
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import errno
import os
import signal
import socket
import struct
import sys
import threading
import time
from optparse import OptionParser

from fprime.constants import DATA_ENCODING

try:
    import socketserver
except ImportError:
    import SocketServer as socketserver


__version__ = 0.1
__date__ = "2015-04-03"
__updated__ = "2016-04-07"

# Universal server id global
SERVER = None
LOCK = None
shutdown_event = threading.Event()

FSW_clients = []
GUI_clients = []
FSW_ids = []
GUI_ids = []


def signal_handler(*_):
    print("Ctrl-C received, server shutting down.")
    shutdown_event.set()


def now():
    return time.ctime(time.time())


class ThreadedTCPRequestHandler(socketserver.StreamRequestHandler):
    """
    Derived from original Stable demo during R&TD and adapted
    for use in new FSW gse.py application.

    TCP socket server for commands, log events, and telemetry data.
    Later this will handle other things such as sequence files and parameters.

    Handle is instanced in own thread for each client.

    Registration is done by sending the string "Register <name>".
    Sending a message to destination <name> is done as
    "A5A5 <name> <data>" Note only <data> is sent.
    Any client that sends a "List" comment makes the server display all
    registered clients.
    """

    socketserver.StreamRequestHandler.allow_reuse_address = True
    socketserver.StreamRequestHandler.timeout = 1

    def handle(self):  # on each client connect
        """
        The function that is invoked upon a new client.  This function listens
        for data on the socket.  Packets for now are assumed to be separated
        by a newline.  For each packet, call processPkt.
        """

        self.partial = b""
        self.cmdQueue = []
        self.registered = False
        self.name = b""
        self.id = 0

        # print self.client_address, now()        # show this client's address
        # Read the data from the socket
        data = self.recv(13)

        # Connection was closed by the client
        if not data:
            print("Client exited.")
            return

        # Process the data into the cmdQueue
        self.getCmds(data)

        # Process the cmdQueue
        self.processQueue()

        if self.registered:
            print("Registration complete waiting for message.")
            self.getNewMsg()
        else:
            print("Unable to register client.")
            return

        LOCK.acquire()
        del SERVER.dest_obj[self.name]
        if self.name in FSW_clients:
            FSW_clients.remove(self.name)
            FSW_ids.remove(self.id)
        elif self.name in GUI_clients:
            GUI_clients.remove(self.name)
            GUI_ids.remove(self.id)
        LOCK.release()

        print(f"Closed {self.name.decode(DATA_ENCODING)} connection.")
        self.registered = False
        self.request.close()

    def getCmds(self, inputString, end_of_command=b"\n"):
        """
        Build a command from partial or full socket input
        """
        commands = inputString.split(end_of_command)
        if len(self.partial):
            commands[0] = self.partial + commands[0]
            self.partial = b""
        if len(commands[-1]):
            self.partial = commands[-1]
        self.cmdQueue.extend(commands[:-1])

    def processQueue(self):
        for cmd in self.cmdQueue:
            self.processRegistration(cmd)
        self.cmdQueue = []

    def processRegistration(self, cmd):

        params = cmd.split()
        process_id = 0

        if params[0] == b"Register":
            LOCK.acquire()
            name = params[1]
            if b"FSW" in name:
                if FSW_clients:
                    process_id = sorted(FSW_ids)[-1] + 1

                name = params[1] + b"_" + bytes(process_id)
                FSW_clients.append(name)
                FSW_ids.append(process_id)
            elif b"GUI" in name:
                if GUI_clients:
                    process_id = sorted(GUI_ids)[-1] + 1

                name = params[1] + b"_" + bytes(process_id)
                GUI_clients.append(name)
                GUI_ids.append(process_id)

            SERVER.dest_obj[name] = DestObj(name, self.request)
            LOCK.release()

            self.registered = True
            self.name = name
            self.id = process_id
            print(f"Registered client {self.name.decode(DATA_ENCODING)}")

    #################################################
    # New Routines to process the command messages
    #################################################
    def getNewMsg(self):
        """
        After registration wait for an incoming message
        The first part must always be an "A5A5 " or a "List "
        """

        # Loop while the connected client has packets to send/receive
        while not shutdown_event.is_set():
            # Read the header data from the socket either A5A5 or List
            header = self.readHeader()

            # If the received header is an empty string, connection closed, exit loop
            if not header:
                break

            if header == b"Quit":
                LOCK.acquire()
                print("Quit received!")
                SERVER.dest_obj[self.name].put(struct.pack(">I", 0xA5A5A5A5))
                shutdown_event.set()
                time.sleep(1)
                print("Quit processed!")
                SERVER.shutdown()
                SERVER.server_close()
                LOCK.release()
                break

            # Got the header data so read the data of the message here...
            data = self.readData(header)

            # Process and send the packet of the message here...
            self.processNewPkt(header, data)

    def recv(self, l):
        """
        Read l bytes from socket.
        """
        chunk = b""
        msg = b""
        n = 0
        while l > n:
            try:
                chunk = self.request.recv(l - n)
                if chunk == b"":
                    print("read data from socket is empty!")
                    return b""
                msg = msg + chunk
                n = len(msg)
            except socket.timeout:
                if shutdown_event.is_set():
                    print("socket timed out and shutdown is requested")
                    return b"Quit\n"
                continue
            except OSError as err:
                if err.errno == errno.ECONNRESET:
                    print(
                        f"Socket error {str(err.errno)} (Connection reset by peer) occurred on recv()."
                    )
                else:
                    print(f"Socket error {str(err.errno)} occurred on recv().")
        return msg

    def readHeader(self):
        """
        Read the 9 byte header (e.g. "A5A5 GUI " or "A5A5 FSW "),
        or just read the "List\n" command.
        """
        header = self.recv(5)

        if len(header) == 0:
            print(f"Header information is empty, client {self.name.decode(DATA_ENCODING)} exiting.")
            return header
        if header == b"List\n":
            return b"List"
        if header == b"Quit\n":
            return b"Quit"
        if header[:-1] == b"A5A5":
            header2 = self.recv(4)
            return header + header2
        return

    def readData(self, header):
        """
        Read the data part of the message sent to either GUI or FSW.
        GUI receives telemetry.
        FSW receives commands of various lengths.
        """
        data = b""
        if header in [b"List", b"Quit"]:
            return b""
        dst = header.split(b" ")[1].strip(b" ")
        if dst == b"FSW":
            # Read variable length command data here...
            desc = self.recv(4)
            sizeb = self.recv(4)
            size = struct.unpack(">I", sizeb)[0]
            data = desc + sizeb + self.recv(size)
        elif dst == b"GUI":
            # Read telemetry data here...
            tlm_packet_size = self.recv(4)
            size = struct.unpack(">I", tlm_packet_size)[0]
            data = tlm_packet_size + self.recv(size)

        else:
            msg = f"unrecognized client {dst.decode(DATA_ENCODING)}"
            raise RuntimeError(msg)
        return data

    def processNewPkt(self, header, data):
        """
        Process a single command here header and data here.
        The command must always start with A5A5 except if it is a List.
        Once the entire header tstring is processed send it on queue.
        If something goes wrong report and shutdown server.
        """
        dest_list = []

        if header == b"List":
            print("List of registered clients: ")
            LOCK.acquire()
            for d in SERVER.dest_obj:
                print("\t" + SERVER.dest_obj[d].name.decode(DATA_ENCODING))
                reg_client_str = b"List " + SERVER.dest_obj[d].name
                l = len(reg_client_str)
                reg_client_str = struct.pack("i%ds" % l, l, reg_client_str)
                self.request.send(reg_client_str)
            LOCK.release()
            return 0

        # Process data here...
        head, dst = header.strip(b" ").split(b" ")
        if head != b"A5A5":
            raise RuntimeError("Packet missing A5A5 header")
        # print "Received Packet: %s %s...\n" % (head,dst)
        if data == b"":
            print(" Data is empty, returning.")
        if b"GUI" in dst:
            dest_list = GUI_clients
        elif b"FSW" in dst:
            dest_list = FSW_clients
        for dest_elem in dest_list:
            # print "Locking TCP"
            LOCK.acquire()
            if dest_elem in SERVER.dest_obj:
                # Send the message here....
                # print "Sending TCP msg to ", dest_elem

                SERVER.dest_obj[dest_elem].put(data)
            LOCK.release()


class ThreadedUDPRequestHandler(socketserver.BaseRequestHandler):
    """
    Derived from original Stable demo during R&TD and adapted
    for use in new FSW gse.py application.

    TCP socket server for commands, log events, and telemetry data.
    Later this will handle other things such as sequence files and parameters.

    Handle is instanced in own thread for each client.

    Registration is done by sending the string "Register <name>".
    Sending a message to destination <name> is done as
    "A5A5 <name> <data>" Note only <data> is sent.
    Any client that sends a "List" comment makes the server display all
    registered clients.
    """

    socketserver.BaseRequestHandler.allow_reuse_address = True

    def handle(self):  # on each packet
        """
        The function that is invoked when a packet is received.  This function listens
        for data on the socket.  Packets for now are assumed to be separated
        by a newline.  For each packet, call processPkt.
        """

        self.getNewMsg(self.request[0])

    #################################################
    # New Routines to process the command messages
    #################################################
    def getNewMsg(self, packet):
        """
        After registration wait for an incoming message
        The first part must always be an "A5A5 " or a "List "
        """

        # Read the header data from the socket either A5A5 or List
        (header, packet) = self.readHeader(packet)

        # If the received header is an empty string, connection closed, exit loop
        if not header:
            return

        # Got the header data so read the data of the message here...
        data = self.readData(header, packet)

        # Process and send the packet of the message here...
        self.processNewPkt(header, data)

    def readHeader(self, packet):
        """
        Read the 9 byte header (e.g. "A5A5 GUI " or "A5A5 FSW "),
        or just read the "List\n" command.
        """
        header = packet[:4]
        header2 = packet[4:9]
        packet = packet[9:]
        return (header + header2, packet)

    def readData(self, header, packet):
        """
        Read the data part of the message sent to either GUI or FSW.
        GUI receives telemetry.
        FSW receives commands of various lengths.
        """
        data = ""
        header.split(b" ")[1].strip(b" ")
        # Read telemetry data here...
        tlm_packet_size = packet[:4]
        size = struct.unpack(">I", tlm_packet_size)[0]
        data = tlm_packet_size + packet[4 : 4 + size]

        return data

    def processNewPkt(self, header, data):
        """
        Process a single command here header and data here.
        The command must always start with A5A5 except if it is a List.
        Once the entire header string is processed send it on queue.
        If something goes wrong report and shutdown server.
        """
        dest_list = []
        # Process data here...
        head, dst = header.strip(b" ").split(b" ")
        if head != b"A5A5":
            raise RuntimeError("Telemetry missing A5A5 header")
        # print "Received Packet: %s %s...\n" % (head,dst)
        if data == "":
            print(" Data is empty, returning.")
        if b"GUI" in dst:
            dest_list = GUI_clients
        else:
            print(f"dest? {dst.decode(DATA_ENCODING)}")
        for dest_elem in dest_list:
            LOCK.acquire()
            if dest_elem in SERVER.dest_obj:
                # Send the message here....
                # print "Sending UDP msg to ", dest_elem

                SERVER.dest_obj[dest_elem].put(data)
            LOCK.release()


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """
    TCP Socket server.

    Keep a dictionary of destination objects containing queues and
    socket id's for writing to destinations.
    """

    dest_obj = {}
    lock_obj = threading.Lock()


class ThreadedUDPServer(socketserver.ThreadingMixIn, socketserver.UDPServer):
    """
    UDP Socket server.
    """


class DestObj:
    """
    Destination object for all clients registered.
    """

    def __init__(self, name, request):
        """
        Constructor
        """
        self.name = name
        self.socket = request
        self.packet = b""

    def put(self, msg):
        """
        Write out the message to the destination socket
        """

        try:
            # print "about to send data to " + self.name
            self.socket.send(msg)
        except OSError as err:
            print(f"Socket error {str(err.errno)} occurred on send().")

    def fileno(self):
        """"""
        return self.socket


def main(argv=None):
    global SERVER, LOCK

    program_name = os.path.basename(sys.argv[0])
    program_license = "Copyright 2015 user_name (California Institute of Technology)                                            \
                ALL RIGHTS RESERVED. U.S. Government Sponsorship acknowledged."
    program_version = "v0.1"
    program_build_date = f"{__updated__}"
    program_version_string = f"%prog {program_version} ({program_build_date})"
    program_longdesc = (
        """"""  # optional - give further explanation about what the program does
    )

    if argv is None:
        argv = sys.argv[1:]

    try:
        parser = OptionParser(
            version=program_version_string,
            epilog=program_longdesc,
            description=program_license,
        )
        parser.add_option(
            "-p",
            "--port",
            dest="port",
            action="store",
            type="int",
            help="Set threaded tcp socket server port [default: %default]",
            default=50050,
        )
        parser.add_option(
            "-i",
            "--host",
            dest="host",
            action="store",
            type="string",
            help="Set threaded tcp socket server ip [default: %default]",
            default="127.0.0.1",
        )

        # process options
        (opts, args) = parser.parse_args(argv)

        HOST = opts.host
        PORT = opts.port
        server = ThreadedTCPServer((HOST, PORT), ThreadedTCPRequestHandler)
        udp_server = ThreadedUDPServer((HOST, PORT), ThreadedUDPRequestHandler)
        # Hopefully this will allow address reuse and server to restart immediately
        server.allow_reuse_address = True
        SERVER = server
        LOCK = server.lock_obj

        print(f"TCP Socket Server listening on host addr {HOST}, port {PORT}")
        # Start a thread with the server -- that thread will then start one
        # more thread for each request
        server_thread = threading.Thread(target=server.serve_forever)
        udp_server_thread = threading.Thread(target=udp_server.serve_forever)
        signal.signal(signal.SIGINT, signal_handler)
        server_thread.daemon = False
        server_thread.start()
        udp_server_thread.daemon = False
        udp_server_thread.start()

        while not shutdown_event.is_set():
            server_thread.join(timeout=5.0)
            udp_server_thread.join(timeout=5.0)

        print("shutdown from main thread")

        SERVER.shutdown()
        SERVER.server_close()
        udp_server.shutdown()
        udp_server.server_close()

        time.sleep(1)

    except Exception as e:
        indent = len(program_name) * " "
        sys.stderr.write(f'{program_name}: {repr(e)}' + "\n")
        sys.stderr.write(indent + "  for help use --help\n")
        return 2


if __name__ == "__main__":
    sys.exit(main())



####
# channels.py:
#
# This file captures the HTML endpoint for the events list. This file contains one endpoint which allows for listing
# events after a certain time-stamp.
#
#  GET /events: list events
#      Input Data: {
#                      "start-time": "YYYY-MM-DDTHH:MM:SS.sss" #Start time for event listing
#                  }
####
from fprime_gds.flask.resource import DictionaryResource, HistoryResourceBase


class ChannelDictionary(DictionaryResource):
    """Channel dictionary shares implementation"""


class ChannelHistory(HistoryResourceBase):
    """
    Resource supplying the history of channels in the system. Includes `get_display_text` postprocessing to add in the
    getter for the display text.
    """



"""
@brief Packet Decoder class used to parse binary packetized telemetry data

Decoders are responsible for taking in serialized data and parsing it into
objects. Decoders receive serialized data (that had a specific descriptor) from
a distributer that it has been registered to. The distributer will send the
binary data after removing any length and descriptor headers.

Example data that would be sent to a decoder that parses events or channels:
    +-------------------+---------------------+------------ - - -
    | ID (2 bytes) | Time Tag (11 bytes) | Data....
    +-------------------+---------------------+------------ - - -

@date Created July 12, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize.time_type import TimeType

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.decoders.ch_decoder import ChDecoder
from fprime_gds.common.utils import config_manager


class PktDecoder(ChDecoder):
    """Decoder class for Packetized Telemetry data"""

    def __init__(self, pkt_name_dict, ch_dict, config=None):
        """
        Constructor

        Args:
            pkt_name_dict: (dict) Packet dictionary. Pkt names should be keys
                                  and PktTemplate objects should be values
            ch_dict: (dict) Channel dictionary. Channel ids should be keys and
                            ChTemplate objects should be values

        Returns:
            An initialized PktDecoder object
        """
        if config is None:
            config = config_manager.ConfigManager.get_instance()
        super().__init__(ch_dict, config)

        self.__dict = pkt_name_dict
        self.id_obj = config.get_type("FwTlmPacketizeIdType")

    def decode_api(self, data):
        """
        Decodes the given data and returns the result.

        This function allows for non-registered code to call the same decoding
        code as is used to parse data passed to the data_callback function.

        Args:
            data: (bytearray) Binary packetized telemetry data to decode

        Returns:
            Parsed version of the input data in the form of a PktData object
            or None if the data is not decodable
        """
        ptr = 0

        # Decode Pkt ID here...
        self.id_obj.deserialize(data, ptr)
        ptr += self.id_obj.getSize()
        pkt_id = self.id_obj.val

        # Decode time...
        pkt_time = TimeType()
        pkt_time.deserialize(data, ptr)
        ptr += pkt_time.getSize()

        if pkt_id not in self.__dict:
            # Don't crash if can't find pkt. Just notify and keep going
            print(
                "Packet decode error: id %d not in dictionary. Time=%s"
                % (pkt_id, pkt_time.to_readable())
            )
            print("Full pkt = \n")
            for i in data:
                print("0x%02x" % ord(i))

            return None

        # Retrieve the template instance for this channel
        pkt_temp = self.__dict[pkt_id]

        ch_temps = pkt_temp.get_ch_list()

        ch_data_objs = []
        for ch_temp in ch_temps:
            val_obj = self.decode_ch_val(data, ptr, ch_temp)
            ptr += val_obj.getMaxSize()
            ch_data_objs.append(ChData(val_obj, pkt_time, ch_temp))

        return ch_data_objs



from __future__ import annotations
from enum import Enum
import inspect
import math
import struct
import typing
from fprime_gds.common.fpy.bytecode.directives import (
    AllocateDirective,
    AndDirective,
    ConstCmdDirective,
    Directive,
    ExitDirective,
    FloatAddDirective,
    FloatDivideDirective,
    FloatExponentDirective,
    FloatFloorDivideDirective,
    FloatModuloDirective,
    FloatMultiplyDirective,
    FloatSubtractDirective,
    GotoDirective,
    MemCompareDirective,
    SignedIntDivideDirective,
    SignedModuloDirective,
    UnsignedIntDivideDirective,
    IntMultiplyDirective,
    FloatLogDirective,
    DiscardDirective,
    StackCmdDirective,
    StorePrmDirective,
    StoreTlmValDirective,
    IfDirective,
    IntAddDirective,
    IntEqualDirective,
    IntNotEqualDirective,
    IntSubtractDirective,
    LoadDirective,
    NoOpDirective,
    NotDirective,
    OrDirective,
    PushValDirective,
    StoreDirective,
    UnsignedLessThanDirective,
    UnsignedLessThanOrEqualDirective,
    UnsignedGreaterThanDirective,
    UnsignedGreaterThanOrEqualDirective,
    SignedGreaterThanDirective,
    SignedGreaterThanOrEqualDirective,
    SignedIntToFloatDirective,
    SignedLessThanDirective,
    SignedLessThanOrEqualDirective,
    UnsignedIntToFloatDirective,
    FloatEqualDirective,
    FloatExtendDirective,
    FloatGreaterThanDirective,
    FloatGreaterThanOrEqualDirective,
    FloatLessThanDirective,
    FloatLessThanOrEqualDirective,
    FloatNotEqualDirective,
    FloatToSignedIntDirective,
    FloatToUnsignedIntDirective,
    FloatTruncateDirective,
    UnsignedModuloDirective,
    WaitAbsDirective,
    WaitRelDirective,
    IntegerZeroExtend16To64Directive,
    IntegerZeroExtend32To64Directive,
    IntegerZeroExtend8To64Directive,
    IntegerSignedExtend16To64Directive,
    IntegerSignedExtend32To64Directive,
    IntegerSignedExtend8To64Directive,
    IntegerTruncate64To16Directive,
    IntegerTruncate64To32Directive,
    IntegerTruncate64To8Directive,
)
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime.common.models.serialize.type_base import BaseType

debug = True

WORD_SIZE = 8
# store return addr and prev stack frame offset in stack frame header
STACK_FRAME_HEADER_SIZE = 2 * WORD_SIZE
MAX_INT64 = 2**63 - 1
MIN_INT64 = -(2**63)
MASK_64_BIT = 2**64 - 1


def overflow_check(val: int) -> int:
    masked_val = val & MASK_64_BIT
    if masked_val > MAX_INT64:
        return masked_val - 2**64
    return masked_val


class DirectiveErrorCode(Enum):
    NO_ERROR = 0
    DIR_OUT_OF_BOUNDS = 1
    TLM_GET_NOT_CONNECTED = 2
    TLM_NOT_FOUND = 3
    TLM_ACCESS_OUT_OF_BOUNDS = 4
    PRM_GET_NOT_CONNECTED = 5
    PRM_NOT_FOUND = 6
    PRM_ACCESS_OUT_OF_BOUNDS = 7
    CMD_SERIALIZE_FAILURE = 8
    DELIBERATE_FAILURE = 9
    STACK_OVERFLOW = 10
    STACK_UNDERFLOW = 11
    INVALID_ARGUMENT = 12
    DIVIDE_BY_ZERO = 13


class FpySequencerModel:

    def __init__(
        self, stack_size=4096, cmd_dict: dict[int, CmdTemplate] = None
    ) -> None:
        self.stack = bytearray()
        self.max_stack_size = stack_size
        self.stack_frame_start = 0
        self.cmd_dict = cmd_dict

        self.dirs: list[Directive] = None
        self.next_dir_idx = 0
        self.tlm_db: dict[int, bytearray] = {}
        self.prm_db: dict[int, bytearray] = {}

    def reset(self):
        self.stack = bytearray()
        self.stack_frame_start = 0

        self.dirs: list[Directive] = None
        self.next_dir_idx = 0
        self.tlm_db: dict[int, bytearray] = {}
        self.prm_db: dict[int, bytearray] = {}

    def dispatch(self, dir: Directive) -> DirectiveErrorCode:
        opcode = dir.opcode
        opcode_name = opcode.name

        handler_fn = None
        for name, func in inspect.getmembers(type(self), inspect.isfunction):
            if not name.startswith("handle"):
                # not a dir handler
                continue
            signature = inspect.signature(func)
            params = list(signature.parameters.values())
            assert len(params) == 2
            assert params[1].annotation is not None
            annotations = typing.get_type_hints(func)
            param_type = annotations[params[1].name]
            if isinstance(dir, param_type):
                handler_fn = func
                break

        if handler_fn is None:
            raise NotImplementedError(opcode_name + " not implemented")

        # otherwise call the handler
        ret = handler_fn(self, dir)
        if ret is None:
            return DirectiveErrorCode.NO_ERROR
        return ret

    def run(self, dirs: list[Directive], tlm: dict[int, bytearray] = None):
        if tlm is None:
            tlm = {}
        self.reset()
        self.dirs = dirs
        self.tlm_db = tlm
        if debug:
            # begin the sequence at dir 0
            print("stack", len(self.stack))
            for byte in range(0, len(self.stack)):

                print(
                    type(self.stack[byte]),
                    end=" ",
                )
            print()
        while self.next_dir_idx < len(self.dirs):
            next_dir = self.dirs[self.next_dir_idx]
            if debug:
                print(f"{self.next_dir_idx}:", next_dir)
            self.next_dir_idx += 1
            result = self.dispatch(next_dir)
            if result != DirectiveErrorCode.NO_ERROR:
                return result
            if debug:
                print("stack", len(self.stack))
                for byte in range(0, len(self.stack)):

                    print(
                        self.stack[byte],
                        end=" ",
                    )
                print()
        return DirectiveErrorCode.NO_ERROR

    def get_int_fmt_str(self, size: int, signed: bool) -> str:
        fmt_char = None
        if size == 1:
            fmt_char = "b"
        elif size == 2:
            fmt_char = "h"
        elif size == 4:
            fmt_char = "i"
        elif size == 8:
            fmt_char = "q"
        else:
            assert False, size
        if not signed:
            fmt_char = fmt_char.upper()

        return ">" + fmt_char

    def push(
        self, val: int | float | bytes | bytearray | bool, signed=True, size=WORD_SIZE
    ):
        if isinstance(val, (bytes, bytearray)):
            self.stack += val
        elif isinstance(val, bool):
            # push a byte onto stack
            self.push(b"\xff" if val else b"\x00")
        elif isinstance(val, float):
            self.push(struct.pack(">d", val))
        else:
            assert isinstance(val, int), val
            # first convert the int into bits
            # have to do some stupid python stuff to deal with negatives
            if val < 0:
                # this should give us the right bit repr for two's complement
                val = val + (1 << (size * 8))

            bits = bin(val)[2:]  # remove "0b"
            # okay now truncate if necessary
            bits = bits[-(size * 8) :]
            self.stack += int(bits, 2).to_bytes(size, byteorder="big", signed=False)

    def pop(self, type=int, signed=True, size=WORD_SIZE) -> int | float | bytearray:
        """pops one word off the stack and interprets it as an int or float, of
        the specified signedness (if applicable) and bit width (if applicable)"""
        value = self.stack[-size:]
        self.stack = self.stack[:-size]
        if type == int:
            fmt_str = self.get_int_fmt_str(size, signed)
            return struct.unpack(fmt_str, value)[0]
        elif type == float:
            if size == 8:
                return struct.unpack(">d", value)[0]
            assert size == 4, size
            return struct.unpack(">f", value)[0]
        elif type == bytes or type == bytearray:
            # compiler knows best. always let them have the last word ;)
            return value
        elif type == bool:
            assert size == 1, size
            return bool(value[0])
        else:
            assert False, type

    def validate_cmd(self, opcode: int, args: bytes) -> bool:
        if self.cmd_dict is None:
            # the user didn't load a command dictionary, just ignore verifying cmds
            return True

        if opcode not in self.cmd_dict:
            print("invalid opcode", opcode)
            return False

        cmd = self.cmd_dict[opcode]
        offset = 0
        for arg in cmd.arguments:
            arg_name, arg_desc, arg_type = arg
            arg_type: type[BaseType]
            arg_value = arg_type()
            arg_value.deserialize(args, offset)
            offset += arg_value.getSize()

        return offset == len(args)

    def handle_allocate_stack(self, dir: AllocateDirective):
        if len(self.stack) + dir.size > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        self.stack += bytearray(0 for i in range(0, dir.size))

    def handle_no_op(self, dir: NoOpDirective):
        pass

    def handle_pop_discard(self, dir: DiscardDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        self.pop(size=dir.size, type=bytes)

    def handle_load(self, dir: LoadDirective):
        if len(self.stack) + dir.size > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        if dir.lvar_offset + self.stack_frame_start + dir.size > len(self.stack):
            return DirectiveErrorCode.STACK_OVERFLOW

        # grab a word beginning at lvar start and put on operand stack
        value = self.stack[
            self.stack_frame_start
            + dir.lvar_offset : (self.stack_frame_start + dir.lvar_offset + dir.size)
        ]
        self.push(value)

    def handle_store(self, dir: StoreDirective):
        if len(self.stack) < dir.size:
            return DirectiveErrorCode.STACK_UNDERFLOW

        if dir.lvar_offset + self.stack_frame_start + dir.size > len(self.stack):
            return DirectiveErrorCode.STACK_OVERFLOW

        # get the last `dir.size` bytes of the stack
        value = self.stack[-dir.size :]
        # remove them from top of stack
        self.stack = self.stack[: -dir.size]
        # put into lvar array at the given offset
        for i in range(0, len(value)):
            self.stack[dir.lvar_offset + self.stack_frame_start + i] = value[i]

    def handle_push_val(self, dir: PushValDirective):
        if len(self.stack) + WORD_SIZE > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW
        self.push(dir.val)

    def handle_wait_rel(self, dir: WaitRelDirective):
        if len(self.stack) < 8:
            return DirectiveErrorCode.STACK_UNDERFLOW

        useconds = self.pop(size=4)
        seconds = self.pop(size=4)

        print("wait rel", seconds, useconds)

    def handle_wait_abs(self, dir: WaitAbsDirective):
        if len(self.stack) < 11:
            return DirectiveErrorCode.STACK_UNDERFLOW
        useconds = self.pop(size=4)
        seconds = self.pop(size=4)
        time_context = self.pop(size=1)
        time_base = self.pop(size=2)

        print("wait abs", time_context, time_base, seconds, useconds)

    def handle_const_cmd(self, dir: ConstCmdDirective):
        print("cmd opcode", dir.cmd_opcode, "args", dir.args)
        # always push CmdResponse.OK
        if not self.validate_cmd(dir.cmd_opcode, dir.args):
            raise RuntimeError("Invalid cmd")
        self.push(0, size=4)

    def handle_stack_cmd(self, dir: StackCmdDirective):
        if len(self.stack) < dir.args_size + 4:
            return DirectiveErrorCode.STACK_UNDERFLOW

        cmd = self.stack[-(dir.args_size + 4) :]
        self.stack = self.stack[: -(dir.args_size + 4)]
        opcode = int.from_bytes(cmd[-4:], signed=False, byteorder="big")

        print(
            "cmd opcode",
            opcode,
            "args",
            cmd[:-4],
        )
        if not self.validate_cmd(opcode, cmd[:-4]):
            raise RuntimeError("Invalid cmd")
        # always push CmdResponse.OK
        self.push(0, size=4)

    def handle_goto(self, dir: GotoDirective):
        if dir.dir_idx > len(self.dirs):
            return DirectiveErrorCode.DIR_OUT_OF_BOUNDS
        self.next_dir_idx = dir.dir_idx

    def handle_if(self, dir: IfDirective):
        if dir.false_goto_dir_index > len(self.dirs):
            return DirectiveErrorCode.DIR_OUT_OF_BOUNDS
        if len(self.stack) < 1:
            return DirectiveErrorCode.STACK_UNDERFLOW
        conditional = self.pop(type=bool, size=1)
        if not conditional:
            self.next_dir_idx = dir.false_goto_dir_index

    def handle_store_tlm_val(self, dir: StoreTlmValDirective):
        whole_value: bytearray = self.tlm_db.get(dir.chan_id, None)
        if whole_value is None:
            return DirectiveErrorCode.TLM_NOT_FOUND

        if (
            self.stack_frame_start + dir.lvar_offset + len(whole_value)
            > self.max_stack_size
        ):
            return DirectiveErrorCode.STACK_OVERFLOW

        self.stack[
            self.stack_frame_start
            + dir.lvar_offset : (
                self.stack_frame_start + dir.lvar_offset + len(whole_value)
            )
        ] = whole_value

    def handle_store_prm(self, dir: StorePrmDirective):
        whole_value: bytearray = self.prm_db.get(dir.prm_id, None)
        if whole_value is None:
            return DirectiveErrorCode.PRM_NOT_FOUND

        if (
            self.stack_frame_start + dir.lvar_offset + len(whole_value)
            > self.max_stack_size
        ):
            return DirectiveErrorCode.STACK_OVERFLOW

        self.stack[
            self.stack_frame_start
            + dir.lvar_offset : (
                self.stack_frame_start + dir.lvar_offset + len(whole_value)
            )
        ] = whole_value

    def handle_or(self, dir: OrDirective):
        if len(self.stack) < 2:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=bool, size=1)
        lhs = self.pop(type=bool, size=1)
        self.push(lhs or rhs)

    def handle_and(self, dir: AndDirective):
        if len(self.stack) < 2:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=bool, size=1)
        lhs = self.pop(type=bool, size=1)
        self.push(lhs and rhs)

    def handle_ieq(self, dir: IntEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        self.push(self.pop() == self.pop())

    def handle_ine(self, dir: IntNotEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        self.push(self.pop() != self.pop())

    def handle_ult(self, dir: UnsignedLessThanDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=False)
        lhs = self.pop(signed=False)
        self.push(lhs < rhs)

    def handle_ule(self, dir: UnsignedLessThanOrEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=False)
        lhs = self.pop(signed=False)
        self.push(lhs <= rhs)

    def handle_ugt(self, dir: UnsignedGreaterThanDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=False)
        lhs = self.pop(signed=False)
        self.push(lhs > rhs)

    def handle_uge(self, dir: UnsignedGreaterThanOrEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=False)
        lhs = self.pop(signed=False)
        self.push(lhs >= rhs)

    def handle_slt(self, dir: SignedLessThanDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(lhs < rhs)

    def handle_sle(self, dir: SignedLessThanOrEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(lhs <= rhs)

    def handle_sgt(self, dir: SignedGreaterThanDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(lhs > rhs)

    def handle_sge(self, dir: SignedGreaterThanOrEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(lhs >= rhs)

    def handle_feq(self, dir: FloatEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        print(rhs, lhs, rhs == lhs)
        self.push(lhs == rhs)

    def handle_fne(self, dir: FloatNotEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        self.push(self.pop(type=float) != self.pop(type=float))

    def handle_flt(self, dir: FloatLessThanDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs < rhs)

    def handle_fle(self, dir: FloatLessThanOrEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs <= rhs)

    def handle_fgt(self, dir: FloatGreaterThanDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs > rhs)

    def handle_fge(self, dir: FloatGreaterThanOrEqualDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs >= rhs)

    def handle_not(self, dir: NotDirective):
        if len(self.stack) < 1:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val = self.pop(type=bool, size=1)
        if val:
            self.push(False)
        else:
            self.push(True)

    def handle_fpext(self, dir: FloatExtendDirective):
        if len(self.stack) < 4:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val_bytes = self.stack[-4:]
        self.stack = self.stack[:-4]
        val_as_float = struct.unpack(">f", val_bytes)[0]

        self.push(val_as_float)

    def handle_siext_8_64(self, dir: IntegerSignedExtend8To64Directive):
        if len(self.stack) < 1:
            return DirectiveErrorCode.STACK_UNDERFLOW
        if len(self.stack) + 7 > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        # pop val off stack
        val = self.pop(type=int, signed=True, size=1)

        self.push(val, signed=True, size=8)

    def handle_siext_16_64(self, dir: IntegerSignedExtend16To64Directive):
        if len(self.stack) < 2:
            return DirectiveErrorCode.STACK_UNDERFLOW
        if len(self.stack) + 6 > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        # pop val off stack
        val = self.pop(type=int, signed=True, size=2)

        self.push(val, signed=True, size=8)

    def handle_siext_32_64(self, dir: IntegerSignedExtend32To64Directive):
        if len(self.stack) < 4:
            return DirectiveErrorCode.STACK_UNDERFLOW
        if len(self.stack) + 4 > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        # pop val off stack
        val = self.pop(type=int, signed=True, size=4)

        self.push(val, signed=True, size=8)

    def handle_ziext_8_64(self, dir: IntegerZeroExtend8To64Directive):
        if len(self.stack) < 1:
            return DirectiveErrorCode.STACK_UNDERFLOW
        if len(self.stack) + 7 > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        # pop val off stack
        val = self.pop(type=int, signed=False, size=1)

        self.push(val, signed=False, size=8)

    def handle_ziext_16_64(self, dir: IntegerZeroExtend16To64Directive):
        if len(self.stack) < 2:
            return DirectiveErrorCode.STACK_UNDERFLOW
        if len(self.stack) + 6 > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        # pop val off stack
        val = self.pop(type=int, signed=False, size=2)

        self.push(val, signed=False, size=8)

    def handle_ziext_32_64(self, dir: IntegerZeroExtend32To64Directive):
        if len(self.stack) < 4:
            return DirectiveErrorCode.STACK_UNDERFLOW
        if len(self.stack) + 4 > self.max_stack_size:
            return DirectiveErrorCode.STACK_OVERFLOW

        # pop val off stack
        val = self.pop(type=int, signed=False, size=4)

        self.push(val, signed=False, size=8)

    def handle_fptrunc(self, dir: FloatTruncateDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val_64 = self.pop(type=float)
        val_32_bytes = struct.pack(">f", val_64)
        self.push(val_32_bytes)

    def handle_itrunc_64_8(self, dir: IntegerTruncate64To8Directive):
        if len(self.stack) < 8:
            return DirectiveErrorCode.STACK_UNDERFLOW

        val = self.pop(type=bytes, size=8)
        val = val[-1:]
        self.push(val)

    def handle_itrunc_64_16(self, dir: IntegerTruncate64To16Directive):
        if len(self.stack) < 8:
            return DirectiveErrorCode.STACK_UNDERFLOW

        val = self.pop(type=bytes, size=8)
        val = val[-2:]
        self.push(val)

    def handle_itrunc_64_32(self, dir: IntegerTruncate64To32Directive):
        if len(self.stack) < 8:
            return DirectiveErrorCode.STACK_UNDERFLOW

        val = self.pop(type=bytes, size=8)
        val = val[-4:]
        self.push(val)

    def handle_fptosi(self, dir: FloatToSignedIntDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val = self.pop(type=float)
        self.push(val)

    def handle_fptoui(self, dir: FloatToUnsignedIntDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val = self.pop(type=float)
        self.push(val)

    def handle_sitofp(self, dir: SignedIntToFloatDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val = self.pop()
        self.push(float(val))

    def handle_uitofp(self, dir: UnsignedIntToFloatDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        val = self.pop(signed=False)
        self.push(float(val))

    def handle_iadd(self, dir: IntAddDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(overflow_check(lhs + rhs))

    def handle_isub(self, dir: IntSubtractDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(overflow_check(lhs - rhs))

    def handle_imul(self, dir: IntMultiplyDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop()
        lhs = self.pop()
        self.push(overflow_check(lhs * rhs))

    def handle_udiv(self, dir: UnsignedIntDivideDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=False)
        lhs = self.pop(signed=False)

        # credit to gemini
        if rhs == 0:
            # C++ behavior for division by zero is undefined.
            return DirectiveErrorCode.DIVIDE_BY_ZERO

        # Perform division, truncating towards zero
        # This is different from Python's // which floors.
        python_quotient = int(lhs / rhs)

        # For division, overflow detection isn't typically done with the mask on the result
        # because the quotient itself is within range
        self.push(python_quotient)

    def handle_sdiv(self, dir: SignedIntDivideDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=True)
        lhs = self.pop(signed=True)

        # credit to gemini
        if rhs == 0:
            # C++ behavior for division by zero is undefined.
            return DirectiveErrorCode.DIVIDE_BY_ZERO

        # Special overflow case: MIN_INT64 / -1
        # This results in MAX_INT64 + 1, which overflows to MIN_INT64 in C++.
        if lhs == MIN_INT64 and rhs == -1:
            self.push(MIN_INT64)  # C++ specific overflow behavior
            return

        python_quotient = int(lhs / rhs)

        self.push(python_quotient)

    def handle_fadd(self, dir: FloatAddDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs + rhs)

    def handle_fsub(self, dir: FloatSubtractDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs - rhs)

    def handle_fmul(self, dir: FloatMultiplyDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs * rhs)

    def handle_fdiv(self, dir: FloatDivideDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs / rhs)

    def handle_smod(self, dir: SignedModuloDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=True)
        lhs = self.pop(signed=True)
        self.push(lhs % rhs, signed=True)

    def handle_umod(self, dir: UnsignedModuloDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(signed=False)
        lhs = self.pop(signed=False)
        self.push(lhs % rhs, signed=False)

    def handle_fmod(self, dir: FloatModuloDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs % rhs)

    def handle_fpow(self, dir: FloatExponentDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs**rhs)

    def handle_log(self, dir: FloatLogDirective):
        if len(self.stack) < WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        operand = self.pop(type=float)
        self.push(math.log(operand))

    def handle_float_floor_div(self, dir: FloatFloorDivideDirective):
        if len(self.stack) < 2 * WORD_SIZE:
            return DirectiveErrorCode.STACK_UNDERFLOW
        rhs = self.pop(type=float)
        lhs = self.pop(type=float)
        self.push(lhs // rhs)

    def handle_exit(self, dir: ExitDirective):
        if len(self.stack) < 1:
            return DirectiveErrorCode.STACK_UNDERFLOW
        success = self.pop(type=bool, size=1)
        if success:
            self.next_dir_idx = len(self.dirs)
        else:
            return DirectiveErrorCode.DELIBERATE_FAILURE

    def handle_memcmp(self, dir: MemCompareDirective):
        if len(self.stack) < dir.size * 2:
            return DirectiveErrorCode.STACK_UNDERFLOW

        rhs = self.pop(type=bytes, size=dir.size)
        lhs = self.pop(type=bytes, size=dir.size)
        print(rhs, lhs)
        self.push(rhs == lhs)



from __future__ import annotations
from abc import ABC
import inspect
from dataclasses import astuple, dataclass, field, fields
from pathlib import Path
import struct
import traceback
import typing
from typing import Union, get_args, get_origin
import zlib

from fprime_gds.common.fpy.error import CompileError

# In Python 3.10+, the `|` operator creates a `types.UnionType`.
# We need to handle this for forward compatibility, but it won't exist in 3.9.
try:
    from types import UnionType

    UNION_TYPES = (Union, UnionType)
except ImportError:
    UNION_TYPES = (Union,)

from fprime_gds.common.fpy.bytecode.directives import (
    StackOpDirective,
    FloatLogDirective,
    Directive,
    ExitDirective,
    WaitAbsDirective,
    WaitRelDirective,
)
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.numerical_types import (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I16Type,
    I32Type,
    I64Type,
    I8Type,
    F32Type,
    F64Type,
    IntegerType,
)
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.bool_type import BoolType
from fprime_gds.common.fpy.parser import (
    AstExpr,
    AstOp,
    AstReference,
    Ast,
    AstAssign,
)
from fprime.common.models.serialize.type_base import BaseType as FppType

MAX_DIRECTIVES_COUNT = 1024
MAX_DIRECTIVE_SIZE = 2048
MAX_STACK_SIZE = 65535

COMPILER_MAX_STRING_SIZE = 128


# this is the "internal" integer type that integer literals have by
# default. it is arbitrary precision
class InternalIntType(IntegerType):
    @classmethod
    def range(cls):
        raise NotImplementedError()

    @staticmethod
    def get_serialize_format():
        raise NotImplementedError()

    @classmethod
    def get_bits(cls):
        raise NotImplementedError()

    @classmethod
    def validate(cls, val):
        if not isinstance(val, int):
            raise RuntimeError()


InternalStringType = StringType.construct_type("InternalStringType", None)


SPECIFIC_NUMERIC_TYPES = (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I16Type,
    I32Type,
    I64Type,
    I8Type,
    F32Type,
    F64Type,
)
SPECIFIC_INTEGER_TYPES = (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I16Type,
    I32Type,
    I64Type,
    I8Type,
)
SIGNED_INTEGER_TYPES = (
    I16Type,
    I32Type,
    I64Type,
    I8Type,
)
UNSIGNED_INTEGER_TYPES = (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
)
SPECIFIC_FLOAT_TYPES = (
    F32Type,
    F64Type,
)


def is_instance_compat(obj, cls):
    """
    A wrapper for isinstance() that correctly handles Union types in Python 3.9+.

    Args:
        obj: The object to check.
        cls: The class, tuple of classes, or Union type to check against.

    Returns:
        True if the object is an instance of the class or any type in the Union.
    """
    origin = get_origin(cls)
    if origin in UNION_TYPES:
        # It's a Union type, so get its arguments.
        # e.g., get_args(Union[int, str]) returns (int, str)
        return isinstance(obj, get_args(cls))

    # It's not a Union, so it's a regular type (like int) or a
    # tuple of types ((int, str)), which isinstance handles natively.
    return isinstance(obj, cls)


# a value of type FppTypeClass is a Python `type` object representing
# the type of an Fprime value
FppTypeClass = type[FppType]


class NothingType(ABC):
    """a type which has no valid values in fprime. used to denote
    a function which doesn't return a value"""

    @classmethod
    def __subclasscheck__(cls, subclass):
        return False


# the `type` object representing the NothingType class
NothingTypeClass = type[NothingType]

@dataclass
class FpyCallable:
    return_type: FppTypeClass | NothingTypeClass
    args: list[tuple[str, FppTypeClass]]


@dataclass
class FpyCmd(FpyCallable):
    cmd: CmdTemplate


@dataclass
class FpyMacro(FpyCallable):
    dir: type[Directive]
    """a function which instantiates the macro given the argument exprs"""


MACROS: dict[str, FpyMacro] = {
    "sleep": FpyMacro(
        NothingType,
        [
            (
                "seconds",
                U32Type,
            ),
            ("microseconds", U32Type),
        ],
        WaitRelDirective,
    ),
    "sleep_until": FpyMacro(NothingType, [("wakeup_time", TimeType)], WaitAbsDirective),
    "exit": FpyMacro(NothingType, [("success", BoolType)], ExitDirective),
    "log": FpyMacro(F64Type, [("operand", F64Type)], FloatLogDirective),
}


@dataclass
class FpyTypeCtor(FpyCallable):
    type: FppTypeClass


@dataclass
class FieldReference:
    """a reference to a field/index of an fprime type"""

    parent: "FpyReference"
    """the qualifier"""
    type: FppTypeClass
    """the fprime type of this reference"""
    offset: int
    """the constant offset in the parent type at which to find this field"""
    name: str = None
    """the name of the field, if applicable"""
    idx: int = None
    """the index of the field, if applicable"""

    def get_from(self, parent_val: FppType) -> FppType:
        """gets the field value from the parent value"""
        assert isinstance(parent_val, self.type)
        assert self.name is not None or self.idx is not None
        value = None
        if self.name is not None:
            if isinstance(parent_val, SerializableType):
                value = parent_val.val[self.name]
            elif isinstance(parent_val, TimeType):
                if self.name == "seconds":
                    value = parent_val.__secs
                elif self.name == "useconds":
                    value = parent_val.__usecs
                elif self.name == "time_base":
                    value = parent_val.__timeBase
                elif self.name == "time_context":
                    value = parent_val.__timeContext
                else:
                    assert False, self.name
            else:
                assert False, parent_val

        else:

            assert isinstance(parent_val, ArrayType), parent_val

            value = parent_val._val[self.idx]

        assert isinstance(value, self.type), (value, self.type)
        return value


# named variables can be tlm chans, prms, callables, or directly referenced consts (usually enums)
@dataclass
class FpyVariable:
    """a mutable, typed value referenced by an unqualified name"""

    type_ref: AstExpr
    """the expression denoting the var's type"""
    declaration: AstAssign
    """the node where this var is declared"""
    type: FppTypeClass | None = None
    """the resolved type of the variable. None if type unsure at the moment"""
    lvar_offset: int | None = None
    """the offset in the lvar array where this var is stored"""


# a scope
FpyScope = dict[str, "FpyReference"]


def create_scope(
    references: dict[str, "FpyReference"],
) -> FpyScope:
    """from a flat dict of strs to references, creates a hierarchical, scoped
    dict. no two leaf nodes may have the same name"""

    base = {}

    for fqn, ref in references.items():
        names_strs = fqn.split(".")

        ns = base
        while len(names_strs) > 1:
            existing_child = ns.get(names_strs[0], None)
            if existing_child is None:
                # this scope is not defined atm
                existing_child = {}
                ns[names_strs[0]] = existing_child

            if not isinstance(existing_child, dict):
                # something else already has this name
                print(
                    f"WARNING: {fqn} is already defined as {existing_child}, tried to redefine it as {ref}"
                )
                break

            ns = existing_child
            names_strs = names_strs[1:]

        if len(names_strs) != 1:
            # broke early. skip this loop
            continue

        # okay, now ns is the complete scope of the attribute
        # i.e. everything up until the last '.'
        name = names_strs[0]

        existing_child = ns.get(name, None)

        if existing_child is not None:
            # uh oh, something already had this name with a diff value
            print(
                f"WARNING: {fqn} is already defined as {existing_child}, tried to redefine it as {ref}"
            )
            continue

        ns[name] = ref

    return base


def union_scope(lhs: FpyScope, rhs: FpyScope) -> FpyScope:
    """returns the two scopes, joined into one. if there is a conflict, chooses lhs over rhs"""
    lhs_keys = set(lhs.keys())
    rhs_keys = set(rhs.keys())
    common_keys = lhs_keys.intersection(rhs_keys)

    only_lhs_keys = lhs_keys.difference(common_keys)
    only_rhs_keys = rhs_keys.difference(common_keys)

    new = FpyScope()

    for key in common_keys:
        if not isinstance(lhs[key], dict) or not isinstance(rhs[key], dict):
            # cannot be merged cleanly. one of the two is not a scope
            print(f"WARNING: {key} is defined as {lhs[key]}, ignoring {rhs[key]}")
            new[key] = lhs[key]
            continue

        new[key] = union_scope(lhs[key], rhs[key])

    for key in only_lhs_keys:
        new[key] = lhs[key]
    for key in only_rhs_keys:
        new[key] = rhs[key]

    return new


FpyReference = typing.Union[
    ChTemplate,
    PrmTemplate,
    FppType,
    FpyCallable,
    FppTypeClass,
    FpyVariable,
    FieldReference,
    dict,  # dict of FpyReference
]
"""some named concept in fpy"""


def get_ref_fpp_type_class(ref: FpyReference) -> FppTypeClass:
    """returns the fprime type of the ref, if it were to be evaluated as an expression"""
    if isinstance(ref, ChTemplate):
        result_type = ref.ch_type_obj
    elif isinstance(ref, PrmTemplate):
        result_type = ref.prm_type_obj
    elif isinstance(ref, FppType):
        # constant value
        result_type = type(ref)
    elif isinstance(ref, FpyCallable):
        # a reference to a callable isn't a type in and of itself
        # it has a return type but you have to call it (with an AstFuncCall)
        # consider making a separate "reference" type
        result_type = NothingType
    elif isinstance(ref, FpyVariable):
        result_type = ref.type
    elif isinstance(ref, type):
        # a reference to a type doesn't have a value, and so doesn't have a type,
        # in and of itself. if this were a function call to the type's ctor then
        # it would have a value and thus a type
        result_type = NothingType
    elif isinstance(ref, FieldReference):
        result_type = ref.type
    elif isinstance(ref, dict):
        # reference to a scope. scopes don't have values
        result_type = NothingType
    else:
        assert False, ref

    return result_type


@dataclass
class CompileState:
    """a collection of input, internal and output state variables and maps"""

    types: FpyScope
    """a scope whose leaf nodes are subclasses of BaseType"""
    callables: FpyScope
    """a scope whose leaf nodes are FpyCallable instances"""
    tlms: FpyScope
    """a scope whose leaf nodes are ChTemplates"""
    prms: FpyScope
    """a scope whose leaf nodes are PrmTemplates"""
    consts: FpyScope
    """a scope whose leaf nodes are instances of subclasses of BaseType"""
    variables: FpyScope = field(default_factory=dict)
    """a scope whose leaf nodes are FpyVariables"""
    runtime_values: FpyScope = None
    """a scope whose leaf nodes are tlms/prms/consts/variables, all of which
    have some value at runtime."""

    def __post_init__(self):
        self.runtime_values = union_scope(
            self.tlms,
            union_scope(self.prms, union_scope(self.consts, self.variables)),
        )

    resolved_references: dict[AstReference, FpyReference] = field(
        default_factory=dict, repr=False
    )
    """reference to its singular resolution"""

    expr_types: dict[AstExpr, FppTypeClass | NothingTypeClass] = field(
        default_factory=dict
    )
    """expr to its fprime type, or nothing type if none"""

    stack_op_directives: dict[AstOp, type[StackOpDirective]] = field(
        default_factory=dict
    )
    """some stack operation to which directive will be emitted for it"""

    type_coercions: dict[AstExpr, FppTypeClass] = field(default_factory=dict)
    """expr to fprime type it must be converted into at runtime"""

    expr_values: dict[AstExpr, FppType | NothingType | None] = field(
        default_factory=dict
    )
    """expr to its fprime value, or nothing if no value, or None if unsure at compile time"""

    directives: dict[Ast, list[Directive] | None] = field(default_factory=dict)
    """a list of code generated by each node, or None/empty list if no directives"""

    node_dir_counts: dict[Ast, int] = field(default_factory=dict)
    """node to the number of directives generated by it"""

    lvar_array_size_bytes: int = 0
    """the size in bytes of the lvar array"""

    start_line_idx: dict[Ast, int] = field(default_factory=dict)
    """the line index at which each node's directives will be included in the output"""

    errors: list[CompileError] = field(default_factory=list)
    """a list of all compile exceptions generated by passes"""

    def err(self, msg, n):
        """adds a compile exception to internal state"""
        self.errors.append(CompileError(msg, n))


class Visitor:
    """visits each class, calling a custom visit function, if one is defined, for each
    node type"""

    def _find_custom_visit_func(self, node: Ast):
        for name, func in inspect.getmembers(type(self), inspect.isfunction):
            if not name.startswith("visit") or name == "visit_default":
                # not a visitor, or the default visit func
                continue
            signature = inspect.signature(func)
            params = list(signature.parameters.values())
            assert len(params) == 3
            assert params[1].annotation is not None
            annotations = typing.get_type_hints(func)
            param_type = annotations[params[1].name]
            if is_instance_compat(node, param_type):
                return func
        else:
            # call the default
            return type(self).visit_default

    def _visit(self, node: Ast, state: CompileState):
        visit_func = self._find_custom_visit_func(node)
        visit_func(self, node, state)

    def visit_default(self, node: Ast, state: CompileState):
        pass

    def run(self, start: Ast, state: CompileState):
        """runs the visitor, starting at the given node, descending depth-first"""

        def _descend(node: Ast):
            if not isinstance(node, Ast):
                return
            children = []
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    children.extend(field_val)
                else:
                    children.append(field_val)

            for child in children:
                if not isinstance(child, Ast):
                    continue
                _descend(child)
                if len(state.errors) != 0:
                    break
                self._visit(child, state)
                if len(state.errors) != 0:
                    break

        _descend(start)
        self._visit(start, state)


class TopDownVisitor(Visitor):

    def run(self, start: Ast, state: CompileState):
        """runs the visitor, starting at the given node, descending breadth-first"""

        def _descend(node: Ast):
            if not isinstance(node, Ast):
                return
            children = []
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    children.extend(field_val)
                else:
                    children.append(field_val)

            for child in children:
                if not isinstance(child, Ast):
                    continue
                self._visit(child, state)
                if len(state.errors) != 0:
                    break
                _descend(child)
                if len(state.errors) != 0:
                    break

        self._visit(start, state)
        _descend(start)


SCHEMA_VERSION = 2

HEADER_FORMAT = "!BBBBBHI"
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)


@dataclass
class Header:
    majorVersion: int
    minorVersion: int
    patchVersion: int
    schemaVersion: int
    argumentCount: int
    statementCount: int
    bodySize: int


FOOTER_FORMAT = "!I"
FOOTER_SIZE = struct.calcsize(FOOTER_FORMAT)


@dataclass
class Footer:
    crc: int


def deserialize_directives(bytes: bytes) -> list[Directive]:
    header = Header(*struct.unpack_from(HEADER_FORMAT, bytes))

    if header.schemaVersion != SCHEMA_VERSION:
        raise RuntimeError(f"Schema version wrong (expected {SCHEMA_VERSION} found {header.schemaVersion})")

    dirs = []
    idx = 0
    offset = HEADER_SIZE
    while idx < header.statementCount:
        offset_and_dir = Directive.deserialize(bytes, offset)
        if offset_and_dir is None:
            raise RuntimeError("Unable to deserialize sequence")
        offset, dir = offset_and_dir
        dirs.append(dir)
        idx += 1

    if offset != len(bytes) - FOOTER_SIZE:
        raise RuntimeError(f"{len(bytes) - FOOTER_SIZE - offset} extra bytes at end of sequence")

    return dirs


def serialize_directives(dirs: list[Directive]) -> tuple[bytes, int]:
    output_bytes = bytes()

    for dir in dirs:
        dir_bytes = dir.serialize()
        if len(dir_bytes) > MAX_DIRECTIVE_SIZE:
            print(CompileError(
                f"Directive {dir} in sequence too large (expected less than {MAX_DIRECTIVE_SIZE}, was {len(dir_bytes)})"
            ))
            exit(1)
        output_bytes += dir_bytes

    header = Header(0, 0, 0, SCHEMA_VERSION, 0, len(dirs), len(output_bytes))
    output_bytes = struct.pack(HEADER_FORMAT, *astuple(header)) + output_bytes

    crc = zlib.crc32(output_bytes) % (1 << 32)
    footer = Footer(crc)
    output_bytes += struct.pack(FOOTER_FORMAT, *astuple(footer))

    return output_bytes, crc



"""F Prime Framer/Deframer Implementation of the CCSDS Space Data Link (TC/TM) Protocols"""
import sys
import struct
import copy

from fprime_gds.common.communication.framing import FramerDeframer
from fprime_gds.plugin.definitions import gds_plugin_implementation

import crc


class SpaceDataLinkFramerDeframer(FramerDeframer):
    """CCSDS Framer/Deframer Implementation for the TC (uplink / framing) and TM (downlink / deframing)
    protocols. This FramerDeframer is used for framing TC data for uplink and deframing TM data for downlink.
    """

    SEQUENCE_NUMBER_MAXIMUM = 256
    TC_HEADER_SIZE = 5
    TM_HEADER_SIZE = 6
    TM_FIXED_FRAME_SIZE = 1024
    TM_TRAILER_SIZE = 2
    TC_TRAILER_SIZE = 2

    # As per CCSDS standard, use CRC-16 CCITT config with init value
    # all 1s and final XOR value of 0x0000
    CRC_CCITT_CONFIG = crc.Configuration(
        width=16,
        polynomial=0x1021,
        init_value=0xFFFF,
        final_xor_value=0x0000,
    )
    CRC_CALCULATOR = crc.Calculator(CRC_CCITT_CONFIG)

    def __init__(self, scid, vcid):
        """ """
        self.scid = scid
        self.vcid = vcid
        self.sequence_number = 0

    def frame(self, data):
        """Frame the supplied data in a TC frame"""
        space_packet_bytes = data
        # CCSDS TC protocol defines the length token as number of bytes in full frame, minus 1
        # so we add to packet size the size of the header and trailer and subtract 1
        length = (
            len(space_packet_bytes) + self.TC_HEADER_SIZE + self.TC_TRAILER_SIZE - 1
        )
        assert length < (pow(2, 10) - 1), "Length too-large for CCSDS format"

        # CCSDS TC Header:
        #  2b -  00 - TF version number
        #  1b - 0/1 - 0 enable FARM checks, 1 bypass FARM
        #  1b - 0/1 - 0 = data (Type-D), 1 = control information (Type-C)
        #  2b -  00 - Reserved
        # 10b -  XX - Spacecraft id
        #  6b -  XX - Virtual Channel ID
        # 10b -  XX - Frame length
        #  8b -  XX - Frame sequence number

        # First 16 bits:
        header_val1_u16 = (
            (0 << 14) |  # TF version number (2 bits)
            (1 << 13) |  # Bypass FARM (1 bit)
            (0 << 12) |  # Type-D (1 bit)
            (0 << 10) |  # Reserved (2 bits)
            ((self.scid & 0x3FF))  # SCID (10 bits)
        )
        # Second 16 bits:
        header_val2_u16 = (
            ((self.vcid & 0x3F) << 10) |  # VCID (6 bits)
            (length & 0x3FF)              # Frame length (10 bits)
        )
        # 8 bit sequence number - always 0 in bypass FARM mode
        header_val3_u8 = 0
        header_bytes = struct.pack(">HHB", header_val1_u16, header_val2_u16, header_val3_u8)
        full_bytes_no_crc = header_bytes + space_packet_bytes
        assert (
            len(header_bytes) == self.TC_HEADER_SIZE
        ), "CCSDS primary header must be 5 octets long"
        assert len(full_bytes_no_crc) == self.TC_HEADER_SIZE + len(
            data
        ), "Malformed packet generated"

        full_bytes = full_bytes_no_crc + struct.pack(
            ">H", self.CRC_CALCULATOR.checksum(full_bytes_no_crc)
        )
        return full_bytes

    def get_sequence_number(self):
        """Get the sequence number and increment - used for TM deframing

        This function will return the current sequence number and then increment the sequence number for the next round.

        Return:
            current sequence number
        """
        sequence = self.sequence_number
        self.sequence_number = (self.sequence_number + 1) % self.SEQUENCE_NUMBER_MAXIMUM
        return sequence

    def deframe(self, data, no_copy=False):
        """Deframe TM frames"""
        discarded = b""
        if not no_copy:
            data = copy.copy(data)
        # Continue until there is not enough data for the header, or until a packet is found (return)
        while len(data) >= self.TM_FIXED_FRAME_SIZE:
            # Read header information
            sc_and_channel_ids = struct.unpack_from(">H", data)
            spacecraft_id = (sc_and_channel_ids[0] & 0x3FF0) >> 4
            virtual_channel_id = (sc_and_channel_ids[0] & 0x000E) >> 1
            # Check if the header is correct with regards to expected spacecraft and VC IDs
            if spacecraft_id != self.scid or virtual_channel_id != self.vcid:
                # If the header is invalid, rotate away a Byte and keep processing
                discarded += data[0:1]
                data = data[1:]
                continue
            # Spacecraft ID and Virtual Channel ID match, so we look at end of frame for CRC
            crc_offset = self.TM_FIXED_FRAME_SIZE - self.TM_TRAILER_SIZE
            transmitted_crc = struct.unpack_from(">H", data, crc_offset)[0]
            if transmitted_crc == self.CRC_CALCULATOR.checksum(data[:crc_offset]):
                # CRC is valid, so we return the deframed data
                deframed_data_len = (
                    self.TM_FIXED_FRAME_SIZE
                    - self.TM_TRAILER_SIZE
                    - self.TM_HEADER_SIZE
                )
                deframed = struct.unpack_from(
                    f">{deframed_data_len}s", data, self.TM_HEADER_SIZE
                )[0]
                # Consume the fixed size frame
                data = data[self.TM_FIXED_FRAME_SIZE :]
                return deframed, data, discarded

            print(
                "[WARNING] Checksum validation failed.",
                file=sys.stderr,
            )
            # Bad checksum, rotate 1 and keep looking for non-garbage
            discarded += data[0:1]
            data = data[1:]
            continue
        return None, data, discarded

    @classmethod
    def get_arguments(cls):
        """Arguments to request from the CLI"""
        return {
            ("--scid",): {
                "type": lambda input_arg: int(input_arg, 0),
                "help": "Spacecraft ID",
                "default": 0x44,
                "required": False,
            },
            ("--vcid",): {
                "type": lambda input_arg: int(input_arg, 0),
                "help": "Virtual channel ID",
                "default": 1,
                "required": False,
            },
        }

    @classmethod
    def check_arguments(cls, scid, vcid):
        """Check arguments from the CLI

        Confirms that the input arguments are valid for this framer/deframer.

        Args:
            scid: spacecraft id
            vcid: virtual channel id
        """
        if scid is None:
            raise TypeError(f"Spacecraft ID not specified")
        if scid < 0:
            raise TypeError(f"Spacecraft ID {scid} is negative")
        if scid > 0x3FF:
            raise TypeError(f"Spacecraft ID {scid} is larger than {0x3FF}")

        if vcid is None:
            raise TypeError(f"Virtual Channel ID not specified")
        if vcid < 0:
            raise TypeError(f"Virtual Channel ID {vcid} is negative")
        if vcid > 0x3F:
            raise TypeError(f"Virtual Channel ID {vcid} is larger than {0x3FF}")

    @classmethod
    def get_name(cls):
        """Name of this implementation provided to CLI"""
        return "raw-space-data-link"

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """Register the MyPlugin plugin"""
        return cls



"""
history.py:

An ordered history that defines what interfaces a history should have within the GDS

:author: koran
"""
import abc

import fprime_gds.common.handlers


class History(fprime_gds.common.handlers.DataHandler):
    """
    An ordered history to support the GDS. Histories are intended to be registered with decoders in
    order to handle incoming objects and store them for retrieval. The default behavior of a
    history is to maintain objects in the order they were enqueued. However, should a sub-history
    want to maintain a different order, this should be made clear to the user and still support the
    calls in this History class.
    """

    @abc.abstractmethod
    def retrieve(self, start=None):
        """
        Retrieve objects from this history. If a starting point is specified, will return a
        sub-list of all objects beginning at starting point in the order to the latest object.

        Args:
            start: a position in the history's order. Start should always be able to be specified
                by an index (int).
        Returns:
            an ordered list of objects
        """
        raise NotImplementedError("This history didn't override the retrieve method.")

    @abc.abstractmethod
    def retrieve_new(self):
        """
        Retrieves an ordered list of objects that have been enqueued since the last call to
        retrieve or retrieve_new.

        Returns:
            an ordered list of objects
        """
        raise NotImplementedError(
            "This history didn't override the retrieve_new method."
        )

    @abc.abstractmethod
    def clear(self, start=None):
        """
        Clears objects from history. A clear that specifies a starting point will clear the history
        such that start becomes the earliest (with respect to the history's order) element in the
        history after objects are removed.

        Args:
            start: a position in the history's order. Start should always be able to be specified
                by an index (int).
        """
        raise NotImplementedError("This history didn't override the clear method.")

    @abc.abstractmethod
    def size(self):
        """
        Accessor for the number of objects in the history
        Returns:
            the number of objects (int)
        """
        raise NotImplementedError("This history didn't override the size method.")



import pytest
from fprime_gds.common.communication.ccsds.space_packet import SpacePacketFramerDeframer
from spacepackets.ccsds.spacepacket import SpacePacketHeader, PacketType, SpacePacket

@pytest.fixture
def framer_deframer():
    return SpacePacketFramerDeframer()

def test_frame_valid_data(framer_deframer):
    """Test framing valid data (if applicable)."""
    # Prefix with 4 bytes corresponding to the APID
    data = b"\x00\x00\x00\x01test_payload"
    framed_data = framer_deframer.frame(data)
    header = SpacePacketHeader.unpack(framed_data)
    assert header.packet_type == PacketType.TC
    assert header.apid == 1
    assert header.data_len == len(data) - 1
    assert header.ccsds_version == 0b000  # Default version for CCSDS packets
    assert header.seq_count == 0

def test_frame_invalid_data(framer_deframer):
    """Test framing valid data with an incorrect DataDescType prefixed."""
    # Prefix with 4 bytes corresponding to the DataDescType
    data = b"\xff\xff\xff\xff" + b"test_payload"
    # Invalid DataDescType, should raise ValueError
    with pytest.raises(ValueError):
        framer_deframer.frame(data)

def test_deframe_valid_packet(framer_deframer):
    """Test deframing a valid space packet."""
    apid = 0x123
    pkt_seq_count = 0x0001
    payload = b"0123456789"
    pkt_data_len = len(payload) - 1
    space_header = SpacePacketHeader(
        packet_type=PacketType.TM,
        apid=apid,
        seq_count=pkt_seq_count,
        data_len=pkt_data_len,
    )
    space_packet_bytes = SpacePacket(space_header, sec_header=None, user_data=payload).pack()

    input_data = b"GARBAGE" + space_packet_bytes + b"TRAILING_GARBAGE"

    deframed, remaining_data, discarded = framer_deframer.deframe(input_data)

    assert deframed == payload
    assert remaining_data == b"TRAILING_GARBAGE"
    assert discarded == b"GARBAGE"

def test_deframe_multiple_packets(framer_deframer):
    """Test deframing multiple concatenated space packets."""
    # First packet
    apid1 = 0x100
    seq1 = 0x000A
    payload1 = b"packet_one_payload"
    pkt_data_len1 = len(payload1) - 1
    header1 = SpacePacketHeader(
        packet_type=PacketType.TM,
        apid=apid1,
        seq_count=seq1,
        data_len=pkt_data_len1,
    )
    packet1 = SpacePacket(header1, sec_header=None, user_data=payload1).pack()

    # Second packet
    apid2 = 0x200
    seq2 = 0x000B
    payload2 = b"another_packet_data"
    pkt_data_len2 = len(payload2) - 1
    header2 = SpacePacketHeader(
        packet_type=PacketType.TM,
        apid=apid2,
        seq_count=seq2,
        data_len=pkt_data_len2,
    )
    packet2 = SpacePacket(header2, sec_header=None, user_data=payload2).pack()

    input_data = packet1 + packet2
    deframed, remaining_data, discarded = framer_deframer.deframe_all(input_data, no_copy=False)

    assert len(deframed) == 2
    assert deframed[0] == payload1
    assert deframed[1] == payload2
    assert remaining_data == b""
    assert discarded == b""

def test_deframe_incomplete_packet(framer_deframer):
    """Test deframing with an incomplete packet at the end."""
    header = SpacePacketHeader(
        packet_type=PacketType.TM,
        apid=0,
        seq_count=0,
        data_len=0,
    )
    packet = SpacePacket(header, sec_header=None, user_data=b"").pack()
    incomplete_packet_bytes = packet[:-1]  # Remove last byte to simulate an incomplete packet
    packets, remaining_data, discarded = framer_deframer.deframe_all(incomplete_packet_bytes, no_copy=False)
    assert len(packets) == 0
    assert remaining_data == incomplete_packet_bytes
    assert discarded == b""

def test_deframe_only_garbage(framer_deframer):
    """Test deframing data that contains no valid packet start."""
    garbage_data = b"this is not a ccsds packet at all"
    packets, remaining_data, discarded = framer_deframer.deframe_all(garbage_data, no_copy=False)
    assert len(packets) == 0
    assert discarded in garbage_data
    assert remaining_data in garbage_data



""" Versioning information for the GDS

This file contains versioning and compatibility definitions for the GDS tool set.
"""

# Currently the GDS is backwards compatible with all dictionaries. 9,9,9 effectively disables this check. It will be
# revisited with a schema version in the next release.
MINIMUM_SUPPORTED_FRAMEWORK_VERSION = (0, 0, 0)
MAXIMUM_SUPPORTED_FRAMEWORK_VERSION = (9, 9, 9)



"""
A file containing utilities for interacting with the Integration Test API
"""

import types
from typing import Any, Callable, Dict, List

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.data_types.sys_data import SysData
from fprime_gds.common.templates.data_template import DataTemplate
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI


def get_upcoming_event(
    test_api: IntegrationTestAPI,
    search_filter: predicates.predicate,
    start_time="NOW",
    timeout: float = 5,
) -> EventData:
    """
    Returns the next event matching the given search filter that occurs after
    this is called. Times out after the given amount if no matching new events
    are found.

    :param test_api: An API instance that will be called to find the next event
    :param search_filter: A predicate each found event is tested against; if
        the event doesn't test "True" against this, we ignore it and keep
        searching
    :param start_time: An optional index or predicate to specify the earliest
        event time to search for
    :param timeout: The maximum time (in seconds) to wait for an event

    :return: The first "EventData" found that passes the filter, or "None" if no
        such event is found within time
    """
    event_filter = predicates.satisfies_all(
        [search_filter, predicates.event_predicate()]
    )
    # Test API only takes integer timeouts
    return test_api.find_history_item(
        event_filter, test_api.get_event_test_history(), start_time, int(timeout)
    )


def get_upcoming_channel(
    test_api: IntegrationTestAPI,
    search_filter: predicates.predicate,
    start_time="NOW",
    timeout: float = 5,
) -> ChData:
    """
    Returns the next telemetry update matching the given search filter that
    occurs after this is called. Times out after the given amount if no
    matching new updates are found.

    :param test_api: An API instance that will be called to find the next update
    :param search_filter: A predicate each found update is tested against; if
        the item doesn't test "True" against this, we ignore it and keep
        searching
    :param start_time: An optional index or predicate to specify the earliest
        update time to search for
    :param timeout: The maximum time (in seconds) to wait for an update

    :return: The first "ChData" found that passes the filter, or "None" if no
        such update is found within time
    """
    channel_filter = predicates.satisfies_all(
        [search_filter, predicates.telemetry_predicate()]
    )
    # Test API only takes integer timeouts
    return test_api.find_history_item(
        channel_filter, test_api.get_telemetry_test_history(), start_time, int(timeout)
    )


def get_item_list(
    item_dictionary: Dict[Any, DataTemplate],
    search_filter: predicates.predicate,
    template_to_data: types.FunctionType,
) -> List[SysData]:
    """
    Returns an ID-sorted list containing all the possible item types that could
    occur on the running F' instance, and information about each one.

    :param item_dictionary: A dictionary full of DataTemplate objects containing
        the events you want to list
    :param search_filter: A Test API predicate used to filter out which items
        to include in the returned list
    :param template_to_data: A callback function accepting a single DataTemplate
        object and returning a SysData object

    :return: An ID sorted list with all the items from the given dictionary
        AFTER being filtered
    """
    # Dictionary has DataTemplates by default, so convert them to SysData so
    # filtering will work properly (several predicates assume SysData types)
    event_data_list = list(
        map(lambda x: template_to_data(x[1]), item_dictionary.items())
    )

    # Filter by using the given predicate on the event values
    event_data_list = list(filter(search_filter, event_data_list))
    event_data_list.sort(key=lambda x: x.get_id())
    return event_data_list


def repeat_until_interrupt(func: Callable, *args):
    """
    Continues to call the input function with the given arguments until the
    user interrupts it.

    :param func: The function you want to call repeatedly. If the function
        returns anything, it MUST return a new, updated tuple of the arguments
        passed into it in the same order, which will be used as the new
        arguments in the next iteration. This is done to allow for persistent
        state between iterations; if needed, create a wrapper for your original
        function to do this. If the function does NOT return anything, the
        original input arguments will continue to be used
    :param args: All keyword arguments you want to pass into "func"
    """
    try:
        while True:
            new_args = func(*args)  # lgtm [py/call/wrong-arguments]
            if new_args:
                args = new_args
    except KeyboardInterrupt:
        pass